// ###############################################################################
// Common Framework for Computer Graphics Courses at FI MUNI.
//
// Copyright (c) 2021-2022 Visitlab (https://visitlab.fi.muni.cz)
// All rights reserved.
// ################################################################################

#include "application.hpp"
#include "data.hpp"

// ----------------------------------------------------------------------------
// Constructors & Destructors
// ----------------------------------------------------------------------------

Application::Application(int initial_width, int initial_height, std::vector<std::string> arguments) : PV112Application(initial_width, initial_height, arguments) {

    glCreateVertexArrays(1, &vao_empty);

    // Task 3.3: Create Vertex Buffer Objects (VBO). Copy vertex data found in data.hpp to VBOs.
    
    //zalozim buffer
    glCreateBuffers(1, &buffer_positions_id);
    glCreateBuffers(1, &buffer_colors_id);

    //nahravam data do jednotlivych bufferu
    glNamedBufferStorage(buffer_positions_id, 12*3*sizeof(float), diamond_positions, 0);
    glNamedBufferStorage(buffer_colors_id,    12*3*sizeof(float), diamond_colors,    0);

    //zakladam VAO pro diamond, a nize ho budu plnit (resp. do nej zapojovat buffery, attributy, atd.)
    glCreateVertexArrays(1, &vao_diamond);

    //definuji "nacitac" - aka posunovac pointeru po bufferu
    const int binder_for_position = 8;     //moje vlastni cislovani
    const int attribute_for_position = 3;  //location ve vertexovym shaderu - musi pasovat se shadery

    const int binder_for_color = 1;
    const int attribute_for_color = 1;
    glVertexArrayVertexBuffer(vao_diamond, binder_for_position, buffer_positions_id, 0, 3 * sizeof(float));
    glVertexArrayVertexBuffer(vao_diamond, binder_for_color,    buffer_colors_id,    0, 3 * sizeof(float));

    //definuji jak se ma cist a vypadat jeden attribuit
    glEnableVertexArrayAttrib(vao_diamond, attribute_for_position);
    glVertexArrayAttribFormat(vao_diamond, attribute_for_position, 3, GL_FLOAT, GL_FALSE, 0);
    //tento atribut se spoji se svym "nacitacem"
    glVertexArrayAttribBinding(vao_diamond, attribute_for_position, binder_for_position);

    glEnableVertexArrayAttrib(vao_diamond, attribute_for_color);
    glVertexArrayAttribFormat(vao_diamond, attribute_for_color, 3, GL_FLOAT, GL_FALSE, 0);
    glVertexArrayAttribBinding(vao_diamond, attribute_for_color, binder_for_color);

    // Task 3.4-3.6: Create Vertex Array Objects (VAO) fetching data from VBOs.

    compile_shaders();
}

Application::~Application() {
    glDeleteVertexArrays(1, &vao_empty);

    // Task 3.3: Delete the Vertex Buffer Objects (VBOs).
    glDeleteVertexArrays(1, &vao_diamond);
    glDeleteBuffers(1, &buffer_positions_id);
    glDeleteBuffers(1, &buffer_colors_id);

    // Task 3.4-3.6: Delete Vertex Array Objects (VAOs).

    Application::delete_shaders();
}

// ----------------------------------------------------------------------------
// Methods
// ----------------------------------------------------------------------------

void Application::delete_shaders() {
    glDeleteProgram(program_triangle);
    glDeleteProgram(program_vao);
}

void Application::compile_shaders() {
    delete_shaders();

    program_triangle = create_program(lecture_shaders_path / "main.vert", lecture_shaders_path / "main.frag");
    program_vao = create_program(lecture_shaders_path / "main_vao.vert", lecture_shaders_path / "main_vao.frag");

    std::cout << "Shaders are reloaded." << std::endl;
}

void Application::update(float delta) {}

void Application::render() {
    glViewport(0, 0, width, height);
    glClear(GL_COLOR_BUFFER_BIT);

    switch (ui_chosen_program) {
    case 0: {
        // Good ol' triangle.
        glUseProgram(program_triangle);
        glBindVertexArray(vao_empty);
        glDrawArrays(GL_TRIANGLES, 0, 3);

        break;
    }
    case 1: {
        // Task 3.1: Draw a square.

        break;
    }
    case 2: {
        // Task 3.2: Draw a square using GL_TRIANGLE_STRIP.

        break;
    }
    case 3: {
        // Task 3.4-3.6: Draw all objects.
        glUseProgram(program_vao);
        glBindVertexArray(vao_diamond);
        glDrawArrays(GL_TRIANGLES, 6, 6);
        break;
    }
    }
}

void Application::render_ui() {
    const float unit = ImGui::GetFontSize();

    ImGui::Begin("Configuration", nullptr, ImGuiWindowFlags_NoDecoration);
    ImGui::SetWindowSize(ImVec2(15 * unit, 20 * unit));
    ImGui::SetWindowPos(ImVec2(2 * unit, 2 * unit));
    ImGui::RadioButton("Triangle", &ui_chosen_program, 0);
    ImGui::RadioButton("Square", &ui_chosen_program, 1);
    ImGui::RadioButton("Square (triangle strip)", &ui_chosen_program, 2);
    ImGui::RadioButton("VAO-based objects", &ui_chosen_program, 3);
    ImGui::End();
}

// ----------------------------------------------------------------------------
// Input Events
// ----------------------------------------------------------------------------

void Application::on_resize(int width, int height) {
    // Calls the default implementation to set the class variables.
    PV112Application::on_resize(width, height);
}

void Application::on_mouse_move(double x, double y) {}

void Application::on_mouse_button(int button, int action, int mods) {}

void Application::on_key_pressed(int key, int scancode, int action, int mods) {
    // Calls default implementation that invokes compile_shaders when 'R key is hit.
    PV112Application::on_key_pressed(key, scancode, action, mods);
}
